# designmatrix.py
#
# Create design matrix
#
# This file is part of HectorP 0.1.2.
#
# HectorP is free software: you can redistribute it and/or modify it under the 
# terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.
#
# HectorP is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with 
# HectorP. If not, see <https://www.gnu.org/licenses/>.
#
#  8/ 2/2019 Machiel Bos, Coimbra
# 29/12/2021 Machiel Bos, Santa Clara
#==============================================================================

import numpy as np
from mpmath import *
import sys
import math
from hectorp.control import Control
from hectorp.control import SingletonMeta
from hectorp.observations import Observations

#==============================================================================
# Class definitions 
#==============================================================================
    
class DesignMatrix(metaclass=SingletonMeta):

    def __init__(self):
        """Define the class variables
        """
        #--- get Control parameters (singleton)
        control = Control()
        try:
            self.verbose = control.params['Verbose']
        except:
            self.verbose = True

        #--- Get observations (singleton)
        self.ts = Observations()

        #--- small number
        EPS = 1.0e-4
    
        #--- Legacy stuff: 
        self.periods = []
        try:
            seasonal_signal  = control.params["seasonalsignal"]
            if seasonal_signal==True:
                self.periods.append(365.25)
        except:
            pass
        try:
            halfseasonal_signal  = control.params["halfseasonalsignal"]
            if halfseasonal_signal==True:
                self.periods.append(365.25/2.0)
        except:
            pass

        #--- How many periods and offsets do we habe?
        try:
            periodic_signals = control.params['periodicsignals']
        except:
            if self.verbose==True:
                print('No extra periodic signals are included.')
            periodic_signals = []

        if np.isscalar(periodic_signals)==True:
            self.periods.append(periodic_signals)
        else:
            for i in range(0,len(periodic_signals)):
                self.periods.append(periodic_signals[i])

        #--- Degree Polynomial
        try:
            degree_polynomial = control.params["DegreePolynomial"]
        except:
            print("No Polynomial degree set, using offset + linear trend")
            degree_polynomial = 1;
        if (degree_polynomial<0 or degree_polynomial>12):
            print("Only polynomial degrees between 0 and 12 are allowed")
            sys.exit()
      
        #--- length of arrays          
        self.n_periods = len(self.periods)
        try:
            estimate_offsets = control.params["estimateoffsets"]
        except:
            estimate_offsets = True
        if estimate_offsets==True:
            self.n_offsets = len(self.ts.offsets)
        else:
            self.n_offsets = 0
        try:
            estimate_postseismic = control.params["estimatepostseismic"]
        except:
            estimate_postseismic = True
        if estimate_postseismic==True:
            self.n_postseismicexp = len(self.ts.postseismicexp)
            self.n_postseismiclog = len(self.ts.postseismiclog)
        else:
            self.n_postseismicexp = 0
            self.n_postseismiclog = 0
        try:
            estimate_sse = control.params["estimateslowslipevent"]
        except:
            estimate_sse = True
        if estimate_sse==True:
            self.n_ssetanh = len(self.ts.ssetanh)
        else:
            self.n_ssetanh = 0
        self.n_degrees = degree_polynomial+1
        
        #--- Number of observations
        m = len(self.ts.data.index)
        if m==0:
            print('Zero length of time series!? am crashing...')
            sys.exit()
       
        #--- Remember time halfway between start and end
        self.th = 0.5*(self.ts.data.index[0] + self.ts.data.index[-1])
 
        n = self.n_degrees + 2*self.n_periods + self.n_offsets + \
                self.n_postseismicexp + self.n_postseismiclog + self.n_ssetanh 
                                           
        self.H = np.zeros((m,n))
        for i in range(0,m):

            #--- Polynomial
            self.H[i,0] = 1.0
            t = self.ts.data.index[i] - self.th
            for j in range(1,self.n_degrees):
                self.H[i,j] = self.H[i,j-1]*t

            #--- Periodic Signal
            for j in range(0,self.n_periods):
                self.H[i,self.n_degrees+2*j+0] = \
                   math.cos(2*math.pi*i*self.ts.sampling_period/self.periods[j])
                self.H[i,self.n_degrees+2*j+1] = \
                   math.sin(2*math.pi*i*self.ts.sampling_period/self.periods[j])

            #--- Offsets
            k = self.n_degrees+2*self.n_periods # use to remember index 
            for j in range(0,self.n_offsets):
                if self.ts.offsets[j]<self.ts.data.index[i]+EPS:
                    self.H[i,k+j] = 1.0

            #--- Post-seismic stuff exp
            k = k + self.n_offsets     # increase each time a bit
            t = self.ts.data.index[i]  # shorter notation
            for j in range(0,self.n_postseismicexp):
                [mjd,T] = self.ts.postseismicexp[j]
                if mjd<t+EPS:
                    self.H[i,k+j] = 1.0 - math.exp(-(t-mjd)/T)

            #--- Post-seismic stuff log
            k = k + self.n_postseismicexp  # add previous n_postseismicexp
            for j in range(0,self.n_postseismiclog):
                [mjd,T] = self.ts.postseismiclog[j]
                if mjd<t+EPS:
                    self.H[i,k+j] = math.log(1.0 + (t-mjd)/T)

            #--- Slow slip event, tanh
            k = k + self.n_postseismiclog  # add previous n_postseismiclog
            for j in range(0,self.n_ssetanh):
                [mjd,T] = self.ts.ssetanh[j]
                if mjd<t+EPS:
                    self.H[i,k+j] = 0.5 * ( math.tanh((t-mjd)/T) - 1.0)



    def compute_amp(self,i,theta,error):
        """ Compute amplitude of periodic signal

        Args:
            i (int) : position in array (index)
            theta (array float): array of cos/sin values
            error (array float): array of associated std values

        Returns:
            amp (float) : estimated amplitude
            amp_err (float) : propagated error
        """

        sigma = 0.5*(error[i] + error[i+1])
        nu  = math.sqrt(pow(theta[i],2.0) + pow(theta[i+1],2.0))
        L12 = hyp1f1(-0.5, 1.0, -pow(nu/sigma,2.0)/2.0)
        amp = float(math.sqrt(math.pi/2.0)*sigma*L12)
        amp_err = float(math.sqrt(2.0*pow(sigma,2.0) + pow(nu,2.0) - \
                  math.pi*pow(sigma,2.0)/2.0*pow(L12,2.0)))

        return [amp,amp_err]



    def compute_pha(self,i,theta,error):
        """ Use Monte Carlo to estimate mean phase and std

        Args:
            i (int) : position in array (index)
            theta (array float): array of cos/sin values
            error (array float): array of associated std values

        Returns:
            amp (float) : estimated amplitude
            amp_err (float) : propagated error
        """

        #--- constant
        deg = 180.0/math.pi

        #--- Mean error of cosine and sine
        sigma = 0.5*(error[i] + error[i+1])
        
        #--- store estimated phase-lag in vector v
        n = 10000
        v = np.empty([n])

        rng = np.random.default_rng()

        x = theta[i+0] + sigma*rng.standard_normal(n)
        y = theta[i+1] + sigma*rng.standard_normal(n)
        v = np.arctan2(y,x)

        pha = v.mean()*deg
        pha_err = v.std()*deg

        return [pha,pha_err]



    def show_results(self,output,theta,error):
        """ Show results from least-squares on screen and save to json-dict

        Args:
            output (dictionary): where the estimate values are saved (json)
            theta (float array) : least-squares estimated parameters
            error (float array) : STD of estimated parameters
        """

        control = Control()
        unit = control.params['PhysicalUnit']
        if self.ts.ts_format=='mom':
            ds = 365.25
            timeunit = 'yr'
        else:
            ds = 3600.0
            timeunit = 'h'

        if self.verbose==True:
            print("bias : {0:.3f} +/- {1:.3f} (at {2:.2f})".\
                                    format(theta[0],error[0],self.th))
							    
            if self.n_degrees>1:
                print("trend: {0:.3f} +/- {1:.3f} {2:s}/{3:s}".\
                             format(ds*theta[1],ds*error[1],unit,timeunit))
            if self.n_degrees>2:
                print("quadratic (half acceleration):" + \
                        "{0:.3f} +/- {1:.3f} {2:s}/{3:s}^2".\
                            format(ds*ds*theta[2],ds*ds*error[2],unit,timeunit))
            for j in range(3,self.n_degrees):
                print("degree {0:d}: {1:.3f} +/- {2:.3f} {3:s}/{4:s}^{0:d}".\
                  format(j,pow(ds,j)*theta[j],pow(ds,j)*error[j],unit,timeunit))
            i = self.n_degrees
            for j in range(0,len(self.periods)):
                [amp,amp_err] = self.compute_amp(i,theta,error)
                [pha,pha_err] = self.compute_pha(i,theta,error)

                print("cos {0:8.3f} : {1:.3f} +/- {2:.3f} {3:s}".format(\
			                  self.periods[j],theta[i],error[i],unit))
                i += 1
                print("sin {0:8.3f} : {1:.3f} +/- {2:.3f} {3:s}".format(\
			                  self.periods[j],theta[i],error[i],unit))
                i += 1
                print("amp {0:8.3f} : {1:.3f} +/- {2:.3f} {3:s}".format(\
			                                 self.periods[j],amp,amp_err,unit))
                print("pha {0:8.3f} : {1:.3f} +/- {2:.3f} degrees".format(\
			                                      self.periods[j],pha,pha_err))
            for j in range(0,self.n_offsets):
                print("offset at {0:10.4f} : {1:7.2f} +/- {2:5.2f} {3:s}".\
			                 format(self.ts.offsets[j],theta[i],error[i],unit))
                i += 1
            for j in range(0,self.n_postseismicexp):
                [mjd,T] = self.ts.postseismicexp[j]
                print('exp relaxation at {0:10.4f} (T={1:8.2f}) : '.format(\
                        mjd,T) + '{0:7.2f} +/- {1:5.2f} {2:s}'.format(theta[i],\
                                                                error[i],unit))
                i += 1
            for j in range(0,self.n_postseismiclog):
                [mjd,T] = self.ts.postseismiclog[j]
                print('log relaxation at {0:10.4f} (T={1:8.2f}) : '.format(\
                        mjd,T) + '{0:7.2f} +/- {1:5.2f} {2:s}'.format(theta[i],\
                                                                error[i],unit))
                i += 1
            for j in range(0,self.n_ssetanh):
                [mjd,T] = self.ts.ssetanh[j]
                print('tanh sse at {0:10.4f} (T={1:8.2f}) : '.format(\
                        mjd,T) + '{0:7.2f} +/- {1:5.2f} {2:s}'.format(theta[i],\
                                                                error[i],unit))
                i += 1
		         

        #--- JSON
        if self.n_degrees>1:
            output['trend'] = ds*theta[1]
            output['trend_sigma'] = ds*error[1]
        if self.n_degrees>2:
            output['quadratic'] = ds*ds*theta[2]
            output['quadratic_sigma'] = ds*ds*error[2]
        for j in range(3,self.n_degrees):
            output['degree{0:3}'.format(j)] = pow(ds,j)*theta[j]
            output['degree{0:3}_sigma'.format(j)] = pow(ds,j)*error[j]

        i = self.n_degrees
        for j in range(0,len(self.periods)):
            [amp,amp_err] = self.compute_amp(i,theta,error)
            [pha,pha_err] = self.compute_pha(i,theta,error)
            output["amp_{0:.3f}".format(self.periods[j])]       = amp 
            output["amp_{0:.3f}_sigma".format(self.periods[j])] = amp_err
            output["pha_{0:.3f}".format(self.periods[j])]       = pha 
            output["pha_{0:.3f}_sigma".format(self.periods[j])] = pha_err

            output["cos_{0:.3f}".format(self.periods[j])] = theta[i] 
            output["cos_{0:.3f}_sigma".format(self.periods[j])] = error[i]
            i += 1 
            output["sin_{0:.3f}".format(self.periods[j])] = theta[i] 
            output["sin_{0:.3f}_sigma".format(self.periods[j])] = error[i]
            i += 1 
        output['jump_epochs'] = self.ts.offsets
        output['jump_sizes']  = theta[i:i+self.n_offsets].tolist()
        output['jump_sigmas'] = error[i:i+self.n_offsets].tolist()
        i += self.n_offsets
        if self.n_postseismicexp>0:
            output['postseismicexp_epochs'] = self.ts.postseismicexp
            output['postseismicexp_sizes']  = \
                                    theta[i:i+self.n_postseismicexp].tolist()
            output['postseismicexp_sigmas'] = \
                                    error[i:i+self.n_postseismicexp].tolist()
        i += self.n_postseismicexp
        if self.n_postseismiclog>0:
            output['postseismiclog_epochs'] = self.ts.postseismiclog
            output['postseismiclog_sizes']  = \
                                    theta[i:i+self.n_postseismiclog].tolist()
            output['postseismiclog_sigmas'] = \
                                    error[i:i+self.n_postseismiclog].tolist()
        i += self.n_postseismiclog
        if self.n_ssetanh>0:
            output['ssetanh_epochs'] = self.ts.ssetanh
            output['ssetanh_sizes']  = theta[i:i+self.n_ssetanh].tolist()
            output['ssetanh_sigmas'] = error[i:i+self.n_ssetanh].tolist()
        output['PhysicalUnit'] = unit


    def add_mod(self,theta):
        """ Compute xhat and add it to the Panda Dataframe

        Args:
            theta (array float): contains estimated least-squares parameters

        """

        xhat = self.H @ theta
        self.ts.add_mod(xhat)



# -*- coding: utf-8 -*-
#
# This file is part of HectorP 0.1.2
#
# HectorP is free software: you can redistribute it and/or modify it under the 
# terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.
#
# HectorP is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with 
# HectorP. If not, see <https://www.gnu.org/licenses/>.
#
# 29/12/2021  Machiel Bos, Santa Clara
#==============================================================================

import os
import sys
import math

#==============================================================================
# Subroutines
#==============================================================================


#---------------------
def compute_date(mjd):
#---------------------
    """ Compute the Modified Julian Date

    Args:
        mjd : modified julian date

    Returns:
        year (int)
        month (int)
        day (int)
        hour (int)
        minute (int)
        second (float)
    """

    jul = int(mjd) + 2400001
    l = jul+68569;
    n = 4*l//146097;
    l = l-(146097*n+3)//4;
    i = 4000*(l+1)//1461001;
    l = l-1461*i//4+31;
    j = 80*l//2447;
    k = l-2447*j//80;
    l = j//11;
    j = j+2-12*l;
    i = 100*(n-49)+i+l;

    year  = i
    month = j
    day   = k

    f      = 24.0*(mjd-math.floor(mjd))
    hour   = int(f)
    f      = 60.0*(f - hour)
    minute = int(f)
    second = 60.0*(f - minute)

    return [year,month,day,hour,minute,second]



#--------------------------------------------------
def compute_mjd(year,month,day,hour,minute,second):
#--------------------------------------------------
    """ Compute the Modified Julian Date

    Args:
        year (int)
        month (int)
        day (int)
        hour (int)
        minute (int)
        second (float)

    Returns:
        mjd : modified julian date
    """

    mjd = 367*year - int(7*(year+int((month+9)/12))/4) + int(275*month/9) + \
                                                      day + 1721014 - 2400001

    mjd += (hour + (minute + second/60.0)/60.0)/24.0

    return mjd


#===============================================================================
# Main program
#===============================================================================

def main():

    args = sys.argv[1:]

    if len(args!=6):
        print("Correct input: date2MJD year month day hour minute second")
        sys.exit()
    else:
        year  = int(args[0])
        month = int(args[1])
        day   = int(args[2])
        hour  = int(args[3])
        minute= int(args[4])
        second= float(args[5])
        mjd = compute_mjd(year,month,day,hour,minute,second)
        print("year   : {0:4d}".format(year))
        print("month  : {0:4d}".format(month))
        print("day    : {0:4d}".format(day))
        print("hour   : {0:4d}".format(hour))
        print("minute : {0:4d}".format(minute))
        print("second : {0:f}".format(second))
        print("MJD    : {0:f}".format(mjd))



# -*- coding: utf-8 -*-
#
# This program tests the generalised Schur algorithm to find the 
# Cholesky decomposition of the inverse covariance matrix.
#
#  This script is part of HectorP 0.1.2
#
#  HectorP is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  HectorP is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with HectorP. If not, see <http://www.gnu.org/licenses/>
#
# 21/2/2021 Machiel Bos, Santa Clara
#===============================================================================

import os
import math
import cmath
import time
import sys
import re
import argparse
import numpy as np
from operator import add
from hectorp.powerlaw import Powerlaw
from numpy.linalg import inv
from numpy.polynomial import Polynomial
from math import sin,cos,pi
from itertools import zip_longest
import pyfftw

#===============================================================================
# Subroutines
#===============================================================================


def multiply (a,b):

    #--- Determine which polynomial is longer
    if len(a)>len(b):
        m = len(a)
    else:
        m = len(b)


    #--- Algorithm only works for powers of 2 and zero padding
    n=1
    while n<m:
        n *= 2
    n *= 2

    #--- forward transform to point value
    Fa = pyfftw.interfaces.numpy_fft.rfft (a,n=n)
    Fb = pyfftw.interfaces.numpy_fft.rfft (b,n=n)

    # convolution
    #Fc = [ Fa [ i ] * Fb [ i ] for i in range ( n ) ]
    Fc = Fa* Fb 

    # reverse transform to get back coefficients
    #c = np.fft.irfft ( Fc )
    c = pyfftw.interfaces.numpy_fft.irfft ( Fc )

    m = len(a)+len(b)-1
   
    #c = np.zeros(m)
    #for i in range(0,m):
    #    c[i] = final[i].real
        
    return c[0:m]

   
 
def levinson(t):
    ''' Use Durbin-Levinson algorithm to compute l1 and l2

    Args:
        t : array containing first column of Toeplitz matrix

    Returns:
        l1,l2 : arrays containing Schur polynomials
        delta : scale factor
    '''

    #--- Durbin-Levinson to compute l1 and l2
    m = len(t)
    r = np.zeros(m-1)
    delta = t[0]
    for i in range(0,m-1):
        if i==0:
            gamma = -t[i+1]/delta
        else:
            gamma = -(t[i+1] + np.dot(t[1:i+1],r[0:i]))/delta
            r[1:i+1] = r[0:i] + gamma*r[i-1::-1]

        r[0] = gamma
        delta = t[0] + np.dot(t[1:i+2],r[i::-1])

    #--- create l1 & l2 using r
    l1 = np.zeros(m)
    l2 = np.zeros(m)

    l1[0]   = 1.0
    l1[1:m] = r[m-2::-1]
    l2[1:m] = r[0:m-1]

    return [l1,l2,delta]


def schur(m):
    ''' Basic Fast Schur
    '''

    n = len(m)
    alpha = np.zeros((n,n))
    beta  = np.zeros((n,n))
    gamma = [1.0]*n

    for k in range(1,n):

        alpha[0,k-1] = m[k]
        beta[0,k-1]  = m[k-1]
        
        for j in range(1,k):
            alpha[j,k-j-1] = alpha[j-1,k-j] - gamma[j]*beta[j-1,k-j]
            beta[j,k-j]    = beta[j-1,k-j]  - gamma[j]*alpha[j-1,k-j]
 
        gamma[k] = alpha[k-1,0]/beta[k-1,0]
        beta[k,0] = beta[k-1,0]*(1.0 - pow(gamma[k],2.0))

    beta[0,n-1] = m[n-1]

    return [alpha,beta,gamma]



def schur2(m):
    '''
    '''
    n = len(m)
    t = np.zeros((n,n))
    gamma = [1.0]*n

    t[0,0] = m[0]
    for j in range(1,n):
        t[j,0] = m[j]
        s = m[j] 
        for k in range(1,j):
            t[j,k] = t[j-1,k-1] + gamma[k]*s
            s = s + gamma[k]*t[j-1,k-1]
        gamma[j] = -s/t[j-1,j-1]
        t[j,j] = t[j-1,j-1] * (1 - pow(gamma[j],2.0))

    #--- scale
    #for i in range(0,n):
    #    t[:,i] /= math.sqrt(t[i,i])

    return [t,gamma]


def schur_polynomials2(gamma):
    ''' Compute Schur polynomials of degree n
    '''

    n = len(gamma)

    xi    = Polynomial([0])
    eta   = Polynomial([1])
    xi_p  = Polynomial([0])
    eta_p = Polynomial([1])

    x   = Polynomial([0,1])
    one = Polynomial([1])

    S     = np.matrix([[eta_p,xi],[xi_p,eta]])
    for i in range(1,n):
        T = np.matrix([[x         ,gamma[i]*one],\
                       [x*gamma[i],one]])

        S_new = S @ T

        S = S_new

    eta_p = S[0,0]
    xi_p  = S[1,0]
    print('eta_p=',eta_p.convert().coef)
    print('xi_p=',xi_p.convert().coef)
    print(eta_p.convert().coef[0:-1] + xi_p.convert().coef[1:])


def schur_polynomials(gamma):
    ''' Compute Schur polynomials of degree n
    '''

    n     = len(gamma)
    S     = [[[1],[0]],[[0],[1]]]
    S_new = [[None,None],[None,None]]
    for i in range(1,n):
        T = [[[0,1],[gamma[i]]],\
             [[0,gamma[i]],[1.0]]]

        S_new[0][0] = [sum(j) for j in zip_longest( \
				multiply(S[0][0],T[0][0]), \
				multiply(S[0][1],T[1][0]), fillvalue=0.0)]
        S_new[1][0] = [sum(j) for j in zip_longest( \
				multiply(S[1][0],T[0][0]), \
				multiply(S[1][1],T[1][0]), fillvalue=0.0)]

        #--- Stupid copy() operator needs to be applied to each element
        S[0][0] = S_new[0][0].copy()
        S[1][1] = S_new[0][0][-1:0:-1].copy()
        S[1][0] = S_new[1][0].copy()
        S[0][1] = S_new[1][0][-1:0:-1].copy()
        #print('eta   = ',S[1][1])
        #print('eta_p = ',S[0][0])
        #print('xi    = ',S[0][1])
        #print('xi_p  = ',S[1][0])

    eta_p = S[0][0]
    xi_p  = S[1][0]
    print('psi_n = ',list( map(add,eta_p[0:-1],xi_p[1:]))) 



def generalised_schur(tm,p_0,q_0):
    ''' Generalised Schur Algorithm

    '''

    if tm==1:
        gamma = -p_0[0]/q_0[0]
        delta_tm = 1.0 - pow(gamma,2.0) 
        a_tm = np.array([0.0,1.0])
        c_tm = np.array([0.0,-gamma])
        return [a_tm,c_tm,delta_tm]
    else:
        m = tm//2
        [a_m,c_m,delta_m] = generalised_schur(m,p_0[0:m],q_0[0:m])
 
        #--- Construct b_m and d_m
        d_m = a_m[-1:0:-1].copy() 
        b_m = c_m[-1:0:-1].copy() 

        #--- Multiply polynomials
        #start_time = time.time()
        part1 =       multiply(d_m,p_0) 
        part2 =  -1 * multiply(b_m,q_0)
        #if m>3200:
        #    print("1 - multiply m=",m,', : {0:8.3f} s n'.\
	#			format(float(time.time() - start_time)))
        #start_time = time.time()
        p_m   = [sum(j) for j in zip_longest(part1,part2, fillvalue=0.0)]
        #if m>3200:
        #    print("1 - sum m=",m,', : {0:8.3f} s n'.\
	#			format(float(time.time() - start_time)))
        
        #start_time = time.time()
        part1 = -1 * multiply(c_m,p_0)
        part2 =      multiply(a_m,q_0)
        #if m>3200:
        #    print("2 - multiply m=",m,', : {0:8.3f} s n'.\
	#			format(float(time.time() - start_time)))
        q_m   = [sum(j) for j in zip_longest(part1,part2, fillvalue=0.0)]
        
        [a_mtm,c_mtm,delta_mtm] = generalised_schur(m,p_m[m:tm],q_m[m:tm])

        #--- Multiply polynomials
        part1 = multiply(a_m,a_mtm)
        part2 = multiply(b_m,c_mtm)
        a_tm  = [sum(j) for j in zip_longest(part1,part2, fillvalue=0.0)]
        
        part1 = multiply(c_m,a_mtm)
        part2 = multiply(d_m,c_mtm)
        c_tm  = [sum(j) for j in zip_longest(part1,part2, fillvalue=0.0)]

        delta_tm = delta_m * delta_mtm
  
        return [a_tm,c_tm,delta_tm]


#===============================================================================
# Main program
#===============================================================================

def main():

    nn = 4*1024 + 1
    noise_model = Powerlaw()
    [t,k_new] = noise_model.create_t(nn,0,[-0.8])
    #C = np.zeros((nn,nn))
    #for i in range(0,nn):
    #    for j in range(0,nn-i):
    #        C[i+j,i] = t[j]
    #        C[i,i+j] = t[j]

    #print('\nmatrix C:\n',C)
    #print('\ninv(C):\n',inv(C))
    #U = np.linalg.cholesky(C)
    #print('\nU:\n',U)
    #print('\ninv(U):\n',inv(U))

    #--- Levinson-Durbin
    print('\nLevingson-Durbin')
    start_time = time.time()
    [l1,l2,delta] = levinson(t)
    print("Levinson : {0:12.6f} s n".format(float(time.time() - start_time)))
    print("l1 = ",l1)

    #L1 = np.zeros((nn,nn))
    #L2 = np.zeros((nn,nn))
    #for i in range(0,nn):
    #    for j in range(0,nn-i):
    #        L1[i+j,i] = l1[j]
    #        L2[i+j,i] = l2[j]
   
    #print("delta = ",delta)
    #print('\nL1 = \n',L1) 
    #print('\nL2 = \n',L2) 
    #print("1/delta * (L1*L1' - L2*L2')\n",1/delta*(L1@L1.T-L2@L2.T))


    #--- Schur
    #[alpha,beta,gamma] = schur(t)
    #print('alpha = \n',alpha)
    #print('beta  = \n',beta)
    #print('gamma = \n',gamma)

    #---
    #[tt,gamma] = schur2(t)
    #print('\nU\n',U,'\n')
    #print(tt,gamma)
    #detC = 1.0
    #for i in range(0,len(gamma)):
    #    detC *= beta[i][0]
    #print('detC=',detC,', ',np.linalg.det(C))

    #--- Schur polynomials
    #schur_polynomials(gamma)
    #schur_polynomials2(gamma)

    #--- Generalised Schur
    #p_0 = [0.0]*nn
    #p_0[0:nn-1] = t[1:nn]
    #q_0 = np.array(t)

    #--- remember this one... gamma's are good
    xi  = np.zeros(nn-1)
    p_0 = np.zeros(nn)
    p_0[0:nn-1] = np.array(t[1:])
    q_0 = np.array(t)
    tm  = len(t)
    #print('\n\n\nt = ',t)
    #print('p_0 = ',p_0)
    #print('q_0 = ',q_0)
    start_time = time.time()
    [a_tm,c_tm,delta_tm] = generalised_schur(tm-1,p_0,q_0)
    print("GSA : {0:12.6f} s n".format(float(time.time() - start_time)))
    print('delta_tm = ',delta_tm)
    for i in range(0,nn-1):
        xi[nn-2-i]=a_tm[i] - c_tm[i+1]
    print(xi)

    #--- Show time lapsed


#--- Call main program
if __name__ == "__main__":
    main()





# white.py
#
# Simple class providing the White noise model
#
#
# This file is part of Hector.
#
# Hector is free software: you can redistribute it and/or modify it under the 
# terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.
#
# Hector is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with 
# Hector. If not, see <https://www.gnu.org/licenses/>.
#
# 5/2/2021 Machiel Bos, Santa Clara
#==============================================================================

import numpy as np
from hectorp.control import Control

#==============================================================================
# Subroutines
#==============================================================================


class White:

    def get_Nparam(self):
        """ Return the number of parameters in White noise model
        
        Returns
        -------
        self.Nparam (int) : total number of parameters === 0 - None!
        """

        return 0


        
    def create_t(self,m,k,param):
        """ Create first row of covariance matrix of white noise
    
        Arguments
        ---------
        m (int) : length of time series
        k (int) : index of param
        param (array float) : --- nothing ---
        
        Returns
        -------
        t (row (m,1)) : first row Toeplitz covariance matrix 
        k_new (int)   : shifted index in param array
        """

        #--- Create first row vector of Covariance matrix
        t = np.zeros(m)
        t[0] = 1.0

        return t, k


  
    def penalty(self,k,param):
        """ Computes penalty for white noise

        Arguments
        ---------
        k (int) : index of param
        param (array float) : --- nothing ---
        
        Returns
        -------
        penalty (float)
        """

        penalty = 0.0 
        return penalty



    def show_results(self,output_single,k,noise_params,sigma):
        """ show estimated noiseparameters

        Args:
            output_single (dictionary) : where values for json file are saved
            k (int) : index where we should start reading noise_params
            noise_params (float-array) : fractions + noise model parameters
            sigma (float) : noise amplitude of white noise
        """

        #--- Get some info from other classes
        control = Control()
        try:
            verbose = control.params['Verbose']
        except:
            verbose = True

        if verbose==True:
            unit = control.params['PhysicalUnit']
            print('sigma     = {0:7.4f} {1:s}'.format(sigma,unit))
            print('No noise parameters to show\n')

        output_single['sigma'] = sigma

        return k # no shift in index




import pyfftw
import multiprocessing
import scipy.signal
import scipy.fft
import numpy
from timeit import Timer

a = pyfftw.empty_aligned((128, 64), dtype='complex128')
b = pyfftw.empty_aligned((128, 64), dtype='complex128')

a[:] = numpy.random.randn(128, 64) + 1j*numpy.random.randn(128, 64)
b[:] = numpy.random.randn(128, 64) + 1j*numpy.random.randn(128, 64)

t = Timer(lambda: scipy.signal.fftconvolve(a, b))

print('Time with scipy.fft default backend: %1.3f seconds' %
      t.timeit(number=100))

# Configure PyFFTW to use all cores (the default is single-threaded)
pyfftw.config.NUM_THREADS = multiprocessing.cpu_count()
print('NUM_THREADS=',pyfftw.config.NUM_THREADS)
pyfftw.config.PLANNER_EFFORT = 'FFTW_MEASURE'

# Use the backend pyfftw.interfaces.scipy_fft
with scipy.fft.set_backend(pyfftw.interfaces.scipy_fft):
     # Turn on the cache for optimum performance
     pyfftw.interfaces.cache.enable()

      # We cheat a bit by doing the planning first
     scipy.signal.fftconvolve(a, b)

     print('Time with pyfftw backend installed: %1.3f seconds' %
            t.timeit(number=100))



# -*- coding: utf-8 -*-
#
# msf.py
#
# A simple class that helps to read and write msf-files
#
# This file is part of HectorP 0.1.2.
#
# HectorP is free software: you can redistribute it and/or modify it under the 
# terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.
#
# HectorP is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with 
# HectorP. If not, see <https://www.gnu.org/licenses/>.
#
#  19/05/2023  Machiel Bos, Santa Clara
#==============================================================================

import os
import sys
import zlib, json
from base64 import b64encode, b64decode

#==============================================================================
# Class definition
#==============================================================================

class MSF():
    """Class to handle msf-files
    """
    
    def __init__(self):
        """ Define some class variables """
        
        self.ZIPJSON_KEY = 'base64(zip(o))'



    def read(self,fname):
        """Read msf-file fname and store the data into the Observation class
        
        Args:
            fname (string) : name of file that will be read
        """

        #--- Check if file exists
        if os.path.isfile(fname)==False:
            print('File {0:s} does not exist'.format(fname))
            sys.exit()

        with open(fname, 'r') as fp:
            json_object = json.load(fp)

        j = self.json_unzip(json_object) 

        header = j['header']
        data   = j['data']

        return [header,data]
        
        
        
    def json_zip(self,j):
        """ Convert dictionary into json -> binary -> base64 string

        Args:
            j (dictionary) : data + metadata we want to store
        
        Return
            base64 string
        """

        j = {
            self.ZIPJSON_KEY: base64.b64encode(
                zlib.compress(
                    json.dumps(j).encode('utf-8')
                )
            ).decode('ascii')
        }

        return j



    def json_unzip(self,j, insist=True):
        """ Unpack a dictionary with a binary string into useable dictionary

        Args:
            j : dictionary with binary string
            insist : option to ignore errors
        
        Return:
            dictionary
        """

        try:
            assert (j[self.ZIPJSON_KEY])
            assert (set(j.keys()) == {self.ZIPJSON_KEY})
        except:
            if insist:
                raise RuntimeError("JSON not in the expected format {"\
				    + str(self.ZIPJSON_KEY) + ": zipstring}")
            else:
                return j

        try:
            j = zlib.decompress(base64.b64decode(j[self.ZIPJSON_KEY]))
        except:
            raise RuntimeError("Could not decode/unzip the contents")

        try:
            j = json.loads(j)
        except: 
            raise RuntimeError("Could interpret the unzipped contents")

        return j



    def write(self,fname,header,data):
        """Write the msf data to a file called fname
        
        Args:
            fname (string) : name of file that will be written
            header (dictionary) : metadata
            data (dictionary)   : data
        """

        j = {header,data}
        self.verify_header(header)

        with open(fname, "w") as fp:
            json.dump(self.zip_json(j), fp) 


    def verify_header(header):
        """ Verify if the dictionary has all the required variables

        Args:
            header (dictionary)

        Returns:
            True if passed
        """

        assert 'sampling_period' in header.keys()
        assert 'mjd' in header.keys()
        assert 'column_names' in header.keys()
        assert 'offsets' in header.keys()

        return True



# observations.py
#
# A simple interface that reads and writes mom-files and stores
# them into a Python class 'observations'.
#
# This file is part of HectorP 0.1.2.
#
# HectorP is free software: you can redistribute it and/or modify it under the 
# terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.
#
# HectorP is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with 
# HectorP. If not, see <https://www.gnu.org/licenses/>.
#
#  31/1/2019  Machiel Bos, Santa Clara
#   1/5/2020  David Bugalho
# 29/12/2021  Machiel Bos, Santa Clara
#  7/ 2/2022  Machiel Bos, Santa Clara
# 10/ 7/2022  Machiel Bos, Santa Clara
#==============================================================================

import pandas as pd
import numpy as np
import os
import sys
import math
import re
from hectorp.control import Control
from hectorp.control import SingletonMeta
from hectorp.msf import MSF
from pathlib import Path

#==============================================================================
# Class definition
#==============================================================================

class Observations(metaclass=SingletonMeta):
    """Class to store my time series together with some metadata
    
    Methods
    -------
    momread(fname)
        read mom-file fname and store the data into the mom class
    momwrite(fname)
        write the momdata to a file called fname
    msfread(fname)
        read msf-file fname and store the data into the mom class
    msfwrite(fname)
        write the msfdata to a file called fname
    make_continuous()
        make index regularly spaced + fill gaps with NaN's
    """
    
    def __init__(self):
        """This is my time series class
        
        This constructor defines the time series in pandas DataFrame data,
        list of offsets and the sampling period (unit days)
        
        """

        #--- Get control parameters (singleton)
        control = Control()
        try:
            self.verbose = control.params['Verbose']
        except:
            self.verbose = True

        #--- Scale factor
        try:
            self.scale_factor = float(control.params['ScaleFactor'])
        except:
            self.scale_factor = 1.0

        #--- class variables
        self.data = pd.DataFrame()
        self.offsets = []
        self.postseismicexp = []
        self.postseismiclog = []
        self.ssetanh = []
        self.sampling_period = 0.0
        self.F = None
        self.percentage_gaps = None
        self.m = 0
        self.column_name=''

        self.ZIPJSON_KEY = 'base64(zip(o))'


        #--- Read filename with observations and the directory
        try:
            self.datafile = control.params['DataFile']
            directory = Path(control.params['DataDirectory'])
            fname = str(directory / self.datafile)
        except Exception as e:
            fname = self.datafile = 'None'
            directory = ''

        #--- Which format?
        try:
            self.ts_format = control.params['TS_format']
        except:
            self.ts_format = 'mom'
        if self.ts_format == 'mom':  
            if not fname=='None':
                self.momread(fname)
        elif self.ts_format == 'msf':   
            #--- Are there columns with estimated trajectory models
            try:
                self.use_residuals = control.params['UseResiduals']
            except:
                self.use_residuals = False
            #--- Which column
            try:
                self.column_name = control.params['ColumnName']
            except Exception as e:
                print(e)
                sys.exit()
            if not fname=='None':
                self.msfread(fname)
        else:
            print('Unknown format: {0:s}'.format(self.ts_format))
            sys.exit()

        #--- Inform the user
        if self.verbose==True:
            print("\nFilename                   : {0:s}".format(fname))
            print("TS_format                  : {0:s}".format(self.ts_format))
            print("ScaleFactor                : {0:f}".\
                                                    format(self.scale_factor))
            if self.ts_format == 'msf':   
                print("Column Name                : {0:s}".\
                                             format(self.column_name))
                print("Use Residuals              : {0:}".\
                                                    format(self.use_residuals))
            if not fname=='None':
                print("Number of observations+gaps: {0:d}".format(self.m))
                print("Percentage of gaps         : {0:5.1f}".\
			                           format(self.percentage_gaps))



    def create_dataframe_and_F(self,t,obs,mod,period):
        """ Convert np.arrays into Panda DataFrame and create matrix F

        Args:
            t (np.array): array with MJD or sod
            obs (np.array): array with observations
            mod (np.array): array with modelled values
            period (floatt): sampling period (unit is days or seconds)
        """
        
        #--- Store sampling period in this class
        self.sampling_period = period

        #---- Create pandas DataFrame
        self.data = pd.DataFrame({'obs':np.asarray(obs)}, \
                                              index=np.asarray(t))
        if len(mod)>0:
            self.data['mod']=np.asarray(mod)
            
        #--- Create special missing data matrix F
        self.m = len(self.data.index)
        n = self.data['obs'].isna().sum()
        self.F = np.zeros((self.m,n))
        j=0
        for i in range(0,self.m):
            if np.isnan(self.data.iloc[i,0])==True:
                self.F[i,j]=1.0
                j += 1

        #--- Compute percentage of gaps
        self.percentage_gaps = 100.0 * float(n) /float(self.m)



    def momread(self,fname):
        """Read mom-file fname and store the data into the mom class
        
        Args:
            fname (string) : name of file that will be read
        """
        #--- Constants
        TINY = 1.0e-6

        #--- Check if file exists
        if os.path.isfile(fname)==False:
            print('File {0:s} does not exist'.format(fname))
            sys.exit()
        
        #--- Read the file (header + time series)
        t = []
        obs = []
        mod = []
        mjd_old = 0.0
        with open(fname,'r') as fp:
            for line in fp:
                cols = line.split()
                if line.startswith('#')==True:
                    if len(cols)>3:
                        if cols[1]=='sampling' and cols[2]=='period':
                            self.sampling_period = float(cols[3])
                    if len(cols)>2:
                        if cols[0]=='#' and cols[1]=='offset':
                            self.offsets.append(float(cols[2]))
                        elif cols[0]=='#' and cols[1]=='exp':
                            mjd = float(cols[2])
                            T   = float(cols[3])
                            self.postseismicexp.append([mjd,T])
                        elif cols[0]=='#' and cols[1]=='log':
                            mjd = float(cols[2])
                            T   = float(cols[3])
                            self.postseismiclog.append([mjd,T])
                        elif cols[0]=='#' and cols[1]=='tanh':
                            mjd = float(cols[2])
                            T   = float(cols[3])
                            self.ssetanh.append([mjd,T])
                else:
                    if len(cols)<2 or len(cols)>3:
                        print('Found illegal row: {0:s}'.format(line))
                        sys.exit()
                    
                    mjd = float(cols[0])
                    #--- Fill gaps with NaN's
                    if mjd_old>0.0:
                        while abs(mjd-mjd_old-self.sampling_period)>TINY:
                            mjd_old += self.sampling_period
                            t.append(mjd_old)
                            obs.append(np.NAN)
                            if len(cols)==3:
                                mod.append(float(np.NAN))
                            if mjd_old>mjd-TINY:
                                print('Someting is very wrong here....')
                                print('mjd={0:f}'.format(mjd))
                                sys.exit()
                    t.append(mjd)
                    mjd_old = mjd
                    obs.append(self.scale_factor * float(cols[1]))
                    if len(cols)==3:
                        mod.append(self.scale_factor * float(cols[2]))
        
        self.create_dataframe_and_F(t,obs,mod,self.sampling_period)



    def msfread(self,fname):
        """Read msf-file fname and store the data into the Observation class
        
        Args:
            fname (string) : name of file that will be read
        """

        TINY = 1.0e-7
        msf  = MSF()

        #--- Read header and data
        [header,data] = msf.read(fname)

        #--- Extract required info from header
        self.sampling_period = header['sampling_period']
        column_names = header['column_names']
        if not self.column_name in column_names:
            print('Could not find column {0:s}'.format(self.column_name))
            sys.exit()
        if self.use_residuals==True:
            mod_column_name = 'mod_' + self.column_name
            if not mod_column_name in column_names:
                print('Could not find column {0:s}'.format(mod_column_name))
                sys.exit()
        all_offsets = header[offsets]
        self.offsets = all_offsets[column_name]
             
        #--- Select specified column data
        sod = data['sod']
        if self.use_residuals==True:
            y = data[self.column_name] - data[mod_column_name]
        else:
            y = data[self.column_name]
   
        sod_old = sod[0]
        t = [sod[0]]
        obs = [y[0]]
        for i in range(1,len(sod))
            while abs(sod[i]-sold_old-self.sampling_period)>TINY:
                sod_old += self.sampling_period
                t.append(sod_old)
                obs.append(np.NAN)
                if sod_old>tt-TINY:
                     print('Someting is very wrong here....')
                     print('mjd={0:f}, sod={1:f}'.format(mjd,sod))
                     sys.exit()
                t.append(sod[i])
                sod_old = sod[i]
                obs.append(self.scale_factor * y[i])

        #---- Create pandas DataFrame
        self.create_dataframe_and_F(t,obs,[],self.sampling_period)
        
        
        
    def momwrite(self,fname):
        """Write the momdata to a file called fname
        
        Args:
            fname (string) : name of file that will be written
        """
        #--- Try to open the file for writing
        try:
            fp = open(fname,'w') 
        except IOError: 
           print('Error: File {0:s} cannot be opened for written.'. \
                                                         format(fname))
           sys.exit()
        if self.verbose==True:
            print('--> {0:s}'.format(fname))
        
        #--- Write header
        fp.write('# sampling period {0:f}\n'.format(self.sampling_period))
                
        #--- Write header offsets
        for i in range(0,len(self.offsets)):
            fp.write('# offset {0:10.4f}\n'.format(self.offsets[i]))
        #--- Write header exponential decay after seismic event
        for i in range(0,len(self.postseismicexp)):
            [mjd,T] = self.postseismicexp[i]
            fp.write('# exp {0:10.4f} {1:5.1f}\n'.format(mjd,T))
        #--- Write header logarithmic decay after seismic event
        for i in range(0,len(self.postseismiclog)):
            [mjd,T] = self.postseismiclog[i]
            fp.write('# exp {0:10.4f} {1:5.1f}\n'.format(mjd,T))
        #--- Write header slow slip event
        for i in range(0,len(self.ssetanh)):
            [mjd,T] = self.sshtanh[i]
            fp.write('# tanh {0:10.4f} {1:5.1f}\n'.format(mjd,T))
 
        #--- Write time series
        for i in range(0,len(self.data.index)):
            if not math.isnan(self.data.iloc[i,0])==True:
                fp.write('{0:12.6f} {1:13.6f}'.format(self.data.index[i],\
                                                  self.data.iloc[i,0]))
                if len(self.data.columns)==2:
                    fp.write(' {0:13.6f}\n'.format(self.data.iloc[i,1]))
                else:
                    fp.write('\n')
            
        fp.close()
        


    def msfwrite(self,fname,header=={}):
        """Write the msf data to a file called fname
        
        Args:
            fname (string) : name of file that will be written

        """

        #--- Create instance of msf
        msf = MSF()

        #--- Do some checking
        mod_column_name = ''
        if self.DataFile=='None' and header=={}:
            print('Both DataFile in ctl-file and header are not specified!')
            sys.exit()
        elif not self.DataFile=='None' and not header=={}:
            print('Both DataFile in ctl-file and header are specified!')
            sys.exit()
        elif not self.DataFile=='None'
            [header,data] = msf.read(fname)

        #--- Start with new data dictionary
        data_new = {}
        for column_name in header['column_names']:
            data_new[column_name] = []

        #--- Do we need to add another column?
        if len(self.data.columns)==2:
            mod_column_name = 'mod_' + self.column_name
            if not mod_column_name in header['column_names']:
                header['column_names'].append(mod_column_name)
                data_new[mod_column_name] = []                
        

        #--- Add sod & observations
        for i in range(0,self.m):
            y = self.data.iloc[i,0]
            if not np.isnan(y):
                for j in range(0,len(header['column_names'])):
                    cname = header['column_names'][j]
                    if cname == self.column_name:
                        data_new[cname].append(y)
                    elif cname == mod_column_name:
                        data_new[cname].append(self.data.iloc[i,1])
                    else:
                        data_new[cname].append(data[cname])

        #--- Finally, write header + data_new to file
        msf.write(fname,header,data_new)



    def show_results(self,output):
        """ add info to json-ouput dict
        """
        
        output['N'] = self.m
        output['gap_percentage'] = self.percentage_gaps 
        if self.ts_format=='mom':
            output['TimeUnit'] = 'days'
        else:
            output['TimeUnit'] = 'seconds'



    def add_offset(self,t):
        """ Add time t to list of offsets
        
        Args:
            t (float): modified julian date or second of day of offset
        """

        EPS   = 1.0e-6
        found = False
        i     = 0
        while i<len(self.offsets) and found==False:
            if abs(self.offsets[i]-t)<EPS:
                found = True
            i += 1
        if found==False:
            self.offsets.append(t)



    def set_NaN(self,index):
        """ Set observation at index to NaN and update matrix F

        Args:
            index (int): index of array which needs to be set to NaN
        """

        self.data.iloc[index,0] = np.nan 
        dummy = np.zeros(self.m)
        dummy[index] = 1.0
        self.F = np.c_[ self.F, dummy ] # add another column to F



    def add_mod(self,xhat):
        """ Add estimated model as column in DataFrame

        Args:
            xhat (array float) : estimated model
        """

        self.data['mod']=np.asarray(xhat)



    def write(self,fname):
        """ Select correct subroutine for writing to file

        Args:
            fname (string): complete name of file
        """

        if self.ts_format=='mom':
            self.momwrite(fname)
        elif self.ts_format=='msf':
            self.msfwrite(fname)
        else:
            print('unknown ts_format: {0:s}'.format(self.ts_format))
            sys.exit()



# control.py
#
# This file is part of HectorP 0.1.2
#
# HectorP is free software: you can redistribute it and/or modify it under the 
# terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.
#
# HectorP is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with 
# HectorP. If not, see <https://www.gnu.org/licenses/>.
#
# 29/12/2021  Machiel Bos, Santa Clara
#==============================================================================

import os
import sys

#==============================================================================
# Class definition
#==============================================================================

class SingletonMeta(type):
    """
    The Singleton class can be implemented in different ways in Python. Some
    possible methods include: base class, decorator, metaclass. We will use the
    metaclass because it is best suited for this purpose.
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """
        Possible changes to the value of the `__init__` argument do not affect
        the returned instance.
        """
        if cls not in cls._instances:
            instance = super().__call__(*args, **kwargs)
            cls._instances[cls] = instance
        return cls._instances[cls]



    def clear(cls):
        _ = cls._instances.pop(cls, None)



    def clear_all(*args, **kwargs):
        SingletonMeta._instances = {}



class Control(metaclass=SingletonMeta):
    """Class to store parameters that prescribe how the analysis should be done
    """
   
    def __init__(self, ctl_file):
        """This is my Control class

        Args:
            ctl_file (string) : name of text-file with parameters
        """

        self.params = {}
   
        file_exists = os.path.exists(ctl_file) 
        if file_exists==False:
            print('Cannot open {0:s}'.format(ctl_file))
            sys.exit()
        else:
            with open(ctl_file,'r') as fp:
                for line in fp:
                    cols = line.split()
                    label = cols[0]
                    if cols[1]=='Yes' or cols[1]=='yes':
                        self.params[label] = True
                    elif cols[1]=='No' or cols[1]=='no':
                        self.params[label] = False
                    elif cols[1].isdigit()==True:
                        self.params[label] = int(cols[1])
                    else:
                        if self.is_float(cols[1])==True:
                            if len(cols)==2:
                                self.params[label] = float(cols[1])
                            else:
                                self.params[label] = []
                                for i in range(1,len(cols)):
                                   self.params[label].append(float(cols[i]))
                        else:
                            if len(cols)==2:
                                self.params[label] = cols[1]
                            elif len(cols)>2:
                                self.params[label] = cols[1:]
                            else:
                                print('found label {0:s} but no value!'.\
								format(label))
                                sys.exit()


    def is_float(self,x):
        """ Check if string is float

        Args:
           x (string) : is this a float string?

        Returns:
           True is number is float
        """
        try:
            float(x)
            return True
        except ValueError:
            return False



# ammargrag.py
#
# Implementation of the fast method of Ammar and Grag / Musicus. 
# It provides a quick subroutine to
# perform least-squares given the design matrix H, the observations y and
# the first column of the Toeplitz covariance matrix C.
#
# Equations are taken from Bos et al. (2013), "Fast error analysis of 
# continuous GNSS observations with missing data", Journal of Geodesy,
# DOI 10.1007/s00190-012-0605-0.
#
#
# This file is part of HectorP 0.1.2.
#
# HectorP is free software: you can redistribute it and/or modify it under the 
# terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.
#
# HectorP is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with 
# HectorP. If not, see <https://www.gnu.org/licenses/>.
#
# 11/12/2022  Machiel Bos, Santa Clara
#===============================================================================

import numpy as np
import math
import time
from numpy import fft
from numpy.linalg import inv
import datetime
from hectorp.control import SingletonMeta
from operator import add
from hectorp.powerlaw import Powerlaw
from numpy.linalg import inv
from math import sin,cos,pi
from itertools import zip_longest
import pyfftw


#===============================================================================
# Class definitions
#===============================================================================


class Schur():

    def __init__(self):
        """ Define Class variables
        """

        #pyfftw.config.NUM_THREADS = 4
        #pyfftw.config.NUM_THREADS = multiprocessing.cpu_count()
        #print('NUM_THREADS=',pyfftw.config.NUM_THREADS)
        #pyfftw.interfaces.cache.enable()
        #pyfftw.config.PLANNER_EFFORT = 'FFTW_PATIENT'

        n_max = 19
        self.Fx = [None]*n_max
        self.Fa = [None]*n_max
        self.Fb = [None]*n_max
        self.x  = [None]*n_max
        
        self.rfft_object = [None]*n_max
        self.irfft_object = [None]*n_max

        n = 2
        for i in range(1,n_max):
            print(i,', n=',n)
            self.x[i]  = pyfftw.empty_aligned(n, dtype='float64')
            self.Fx[i] = pyfftw.empty_aligned(n//2+1, dtype='complex128')
            self.Fa[i] = pyfftw.empty_aligned(n//2+1, dtype='complex128')
            self.Fb[i] = pyfftw.empty_aligned(n//2+1, dtype='complex128')
            #self.x[i]  = np.empty(n, dtype='float64')
            #self.Fx[i] = np.empty(n//2+1, dtype='complex128')
            #self.Fa[i] = np.empty(n//2+1, dtype='complex128')
            #self.Fb[i] = np.empty(n//2+1, dtype='complex128')
            
            self.rfft_object[i] = pyfftw.builders.rfft(self.x[i])
            self.irfft_object[i]= pyfftw.builders.irfft(self.Fx[i])


            n *= 2

        self.zeros = np.zeros(n)

        self.multiply = self.multiply_fftw



    def levinson(self,t):
        """ Use Durbin-Levinson algorithm to compute l1 and l2

        This subroutine is just listed here for testing and not
        used in production

        Args:
            t : array containing first column of Toeplitz matrix

        Returns:
            l1,l2 : arrays containing Schur polynomials
            delta : scale factor
        """

        #--- Durbin-Levinson to compute l1 and l2
        m = len(t)
        r = np.zeros(m-1)
        delta = t[0]
        for i in range(0,m-1):
            if i==0:
                gamma = -t[i+1]/delta
            else:
                gamma = -(t[i+1] + np.dot(t[1:i+1],r[0:i]))/delta
                r[1:i+1] = r[0:i] + gamma*r[i-1::-1]

            r[0] = gamma
            delta = t[0] + np.dot(t[1:i+2],r[i::-1])

        #--- create l1 & l2 using r
        l1 = np.zeros(m)
        l2 = np.zeros(m)

        l1[0]   = 1.0
        l1[1:m] = r[m-2::-1]
        l2[1:m] = r[0:m-1]

        return [l1,l2,delta]



    def multiply_fftw(self,a,b):
        """ Multiply 2 polynomials using Numpy FFT

        Args:
            a (array float) : polynomial a0, a1*t, a2*t^2, ...  
            b (array float) : polynomial b0, b1*t, b2*t^2, ...  

        Returns:
            c (array float) : a*b
        """
        #--- Determine which polynomial is longer
        if len(a)>len(b):
            m = len(a)
        else:
            m = len(b)

        #--- Algorithm only works for powers of 2 and zero padding
        n=1
        i=1
        while n<m:
            n *= 2
            i += 1
        n *= 2 #--- These are the extra cells for padding

        #--- forward transform to point value
        #self.x[i] = (ctypes.c_double*len(a)).from_address(id(a))
        self.x[i][0:len(a)] = a[:]
        self.x[i][len(a):n] = self.zeros[len(a):n]
        self.Fa[i][:] = self.rfft_object[i]()

        self.x[i][0:len(b)] = b[:]
        self.x[i][len(b):n] = self.zeros[len(b):n]
        self.Fb[i][:] = self.rfft_object[i]()

        #self.Fx[i][:] = self.Fa[i] * self.Fb[i]
        self.Fx[i][:] = np.multiply(self.Fa[i],self.Fb[i])

        c = self.irfft_object[i]().copy()

        m = len(a)+len(b)-1
        
        return c[0:m]



    def multiply_numpy(self,a,b):
        """ Multiply 2 polynomials using Numpy FFT

        Args:
            a (array float) : polynomial a0, a1*t, a2*t^2, ...  
            b (array float) : polynomial b0, b1*t, b2*t^2, ...  

        Returns:
            c (array float) : a*b
        """

        #--- Determine which polynomial is longer
        if len(a)>len(b):
            m = len(a)
        else:
            m = len(b)

        #--- Algorithm only works for powers of 2 and zero padding
        n=1
        while n<m:
            n *= 2
        n *= 2 #--- These are the extra cells for padding

        #--- forward transform to point value
        Fa = np.fft.rfft (a,n=n) # n=n takes care of zero padding of a[len(a):n]
        Fb = np.fft.rfft (b,n=n)

        #--- convolution
        Fc = Fa * Fb 

        #--- reverse transform to get back coefficients
        c = np.fft.irfft ( Fc )

        m = len(a)+len(b)-1
        
        return c[0:m]



    def generalised_schur(self,tm,p_0,q_0):
        """ Generalised Schur Algorithm

        """

        if tm==1:
            gamma = -p_0[0]/q_0[0]
            delta_tm = 1.0 - pow(gamma,2.0) 
            a_tm = np.array([0.0,1.0])
            c_tm = np.array([0.0,-gamma])
            return [a_tm,c_tm,delta_tm]
        else:
            m = tm//2
            [a_m,c_m,delta_m] = self.generalised_schur(m,p_0[0:m],q_0[0:m])
 
            #--- Construct b_m and d_m
            d_m = a_m[-1:0:-1].copy() 
            b_m = c_m[-1:0:-1].copy() 

            #--- Multiply polynomials
            #start_time = time.time()
            part1 =       self.multiply(d_m,p_0) 
            part2 =  -1 * self.multiply(b_m,q_0)
        
            p_m   = [sum(j) for j in zip_longest(part1,part2, fillvalue=0.0)]
      
            #start_time = time.time()
            part1 = -1 * self.multiply(c_m,p_0)
            part2 =      self.multiply(a_m,q_0)
     
            q_m   = [sum(j) for j in zip_longest(part1,part2, fillvalue=0.0)]
        
            [a_mtm,c_mtm,delta_mtm] = \
                                self.generalised_schur(m,p_m[m:tm],q_m[m:tm])

            #--- Multiply polynomials
            part1 = self.multiply(a_m,a_mtm)
            part2 = self.multiply(b_m,c_mtm)
            a_tm  = [sum(j) for j in zip_longest(part1,part2, fillvalue=0.0)]
        
            part1 = self.multiply(c_m,a_mtm)
            part2 = self.multiply(d_m,c_mtm)
            c_tm  = [sum(j) for j in zip_longest(part1,part2, fillvalue=0.0)]

            delta_tm = delta_m * delta_mtm
  
            return [a_tm,c_tm,delta_tm]


    def test(self):
        """ Perform test computation to check if everything is okay
        """

        #nn = 4*1024 + 1
        nn = 64*1024 + 1
        noise_model = Powerlaw()
        [t,k_new] = noise_model.create_t(nn,0,[-0.8])
  
        #--- Levinson-Durbin
        #print('\nLevingson-Durbin')
        #start_time = time.time()
        #[l1,l2,delta] = self.levinson(t)
        #print("Levinson : {0:12.6f} s n".format(float(time.time() - start_time)))
        #print("l1 = ",l1)

            #--- remember this one... gamma's are good
        xi  = np.zeros(nn-1)
        p_0 = np.zeros(nn)
        p_0[0:nn-1] = np.array(t[1:])
        q_0 = np.array(t)
        tm  = len(t)

        start_time = time.time()
        [a_tm,c_tm,delta_tm] = self.generalised_schur(tm-1,p_0,q_0)
        print("GSA : {0:12.6f} s n".format(float(time.time() - start_time)))
        print('delta_tm = ',delta_tm)
        for i in range(0,nn-1):
            xi[nn-2-i]=a_tm[i] - c_tm[i+1]
        print(xi)



# -*- coding: utf-8 -*-
#
# msfgen.py
#
# Create msf-json files
#
# Example:
# --------
# msfgen -ji meta.json -di imu.dat -o drone_flight1.msf
#
# This program is part of HectorP 0.1.2
#
# HectorP is free software: you can redistribute it and/or modify it under the 
# terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.
#
# HectorP is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with 
# HectorP. If not, see <https://www.gnu.org/licenses/>.
#
# 19/5/2023 Machiel Bos, Santa Clara
#
# (c) Copyright 2023 TeroMovigo, all rights reserved.
#===============================================================================

import os
import sys
import argparse
import numpy as np
import json
from hectorp.msf import MSF

#===============================================================================
# Subroutines
#===============================================================================

def open_file(fname):
    """ Simply test if file exists and open it

    Args:
        fname (string) : name of file to be tested
    """

    try:
        fp = open(fname,'r') 
    except IOError: 
        print('Error: File {0:s} cannot be opened for reading.'.format(fname))
        sys.exit()

    return fp

#===============================================================================
# Main program
#===============================================================================


def main():

    #--- Parse command line arguments in a bit more professional way
    parser = argparse.ArgumentParser(description= 'Create msf-json file')

    #--- List arguments that can be given 
    parser.add_argument('-ji', action='store_true', required=True,
                           dest='fname_header',help='json file with metadata')
    parser.add_argument('-di', action='store_true',required=True,
                           dest='fname_data',help='ASCII data file')
    parser.add_argument('-o', action='store_true', required=True,
                           dest='fname_out',help='output binary json file')

    args = parser.parse_args()

    #--- parse command-line arguments
    fname_header = args.fname_header
    fname_data   = args.fname_data
    fname_out    = args.fname_out

    #--- Create instance of MSF class
    msf = MSF()

    #--- Read metadata
    fp = open_file(fname_header)
    header = json.loads(fp.read()) 
    fp.close()

    #--- Do we have a valid msf header?
    if msf.validate_header() == False:
        sys.exit()

    #--- Already create the arrays we are going to fill
    data = {}
    column_names = header['column_names']
    for i in range(0,len(column_names)):
        data[column_names[i]] = []

    #--- Read the data
    fp = open_file(fname_data) 
    for line in fp:
        if not line.startswith('#'):
            cols = line.split()
            if not len(cols)=len(column_names):
                print('Ooops, number of columns is : {0:d}'.format(len(cols)))
                sys.exit()
            for i in range(0,len(column_names)):
                data[column_names[i]].append(float(cols[i]))

    #--- Write compressed dictionary to file
    msf.write(fname,header,data)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# msfdump.py
#
# Create text file from msf-json file
#
# Example:
# --------
# msfdump -i drone_flight1.msf -o data.txt
#
# This program is part of HectorP 0.1.2
#
# HectorP is free software: you can redistribute it and/or modify it under the 
# terms of the GNU General Public License as published by the Free Software 
# Foundation, either version 3 of the License, or (at your option) any later 
# version.
#
# HectorP is distributed in the hope that it will be useful, but WITHOUT ANY 
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with 
# HectorP. If not, see <https://www.gnu.org/licenses/>.
#
# 19/5/2023 Machiel Bos, Santa Clara
#
# (c) Copyright 2023 TeroMovigo, all rights reserved.
#===============================================================================

import os
import sys
import argparse
import numpy as np
import json
from hectorp.msf import MSF

#===============================================================================
# Main program
#===============================================================================


def main():

    #--- Parse command line arguments in a bit more professional way
    parser = argparse.ArgumentParser(description= 'Dump msf-json file')

    #--- List arguments that can be given 
    parser.add_argument('-i', action='store_true', required=True,
                           dest='fname_in',help='binary json file')
    parser.add_argument('-o', action='store_true', required=True,
                           dest='fname_out',help='output text file')

    args = parser.parse_args()

    #--- parse command-line arguments
    fname_in  = args.fname_in
    fname_out = args.fname_out

    #--- Create instance of MSF class
    msf = MSF()

    #--- Read file
    [header,data] = msf.read(fname_in)

    fp_out = open(fname_out,'w')
    fp_out.write('# sampling period {0:f}\n'.format(header['sampling_period']))
    fp_out.write('#\n# Observations\n# ------------\n')
    j = 1
    for i in range(0,len(header['column_names'])):
        if not header['column_names'][i].startswith('mod_')
            fp_out.write('# {0:d}. {1:s}\n'.format(j,header['column_names'][i]))
            j += 1

    if not j == len(header['column_names']):
        fp_out.write('#\n# Models\n# ------\n')
        for k in range(j,len(header['column_names']):
            fp_out.write('# {0:d}. {1:s}\n'.format(k,header['column_names'][i]))
					       

    if 'Offsets' in header.keys():
        fp_out.write('#\n# Offsets\n# -------\n')
        offsets = header['offsets']
        for cname in offsets.keys():
            i = header['column_names'].index(cname)
            sods = offsets[cname]
            for k in range(0,len(sods)):
                fp_out.write('# {0:d}. {1:f}\n'.format(i,sods[k])

    fp_out.write('#\n#=========================================' + \
                  '=================================================')


    for i in range(0,len(data['sod'])):
        fp_out.write('{0:9.3f}'.format(data['sod'][i]))
        for j in range(0,len(header['column_names'])):
             cname = header['column_names'][j]
             fp_out.write(' {0:11.7f}'.format(data[cname][i]))
        fp_out.write('\n')

    fp_out.close()
    


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#
# This program uses the estimated noise models parameters to predict the 
# trend error for given number of observations.
#
#  This script is part of HectorP 0.1.2
#
#  HectorP is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  HectorP is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with HectorP. If not, see <http://www.gnu.org/licenses/>
#
# 19/8/2022 Machiel Bos, Santa Clara
#===============================================================================

import os
import math
import time
import json
import sys
import numpy as np
from matplotlib import pyplot as plt
from hectorp.control import Control
from hectorp.white import White
from hectorp.ggm import GGM
from hectorp.powerlaw import Powerlaw
from hectorp.varyingannual import VaryingAnnual
from hectorp.ar1 import AR1
from hectorp.ammargrag import AmmarGrag
import argparse
from pathlib import Path

#===============================================================================
# Subroutines
#===============================================================================


#===============================================================================
# Main program
#===============================================================================

def main():

    #--- Parse command line arguments in a bit more professional way
    parser = argparse.ArgumentParser(description= 'Predict trend error')

    #--- List arguments that can be given 
    parser.add_argument('-graph', action='store_true', required=False,
       					help='No graph is shown on screen')
    parser.add_argument('-seasonal', action='store_true',required=False,
       					help='Add yearly signal')
    parser.add_argument('-dt', required=False, default='1', \
                                           dest='dt', help='sampling period')
    parser.add_argument('-t0', required=False, default='730', \
                                           dest='t0', help='t0')
    parser.add_argument('-t1', required=False, default='7300', \
                                           dest='t1', help='t1')
    parser.add_argument('-i', required=False, default='estimatetrend.json', \
                                          dest='fname', help='json filename')

    args = parser.parse_args()

    #--- parse command-line arguments
    graph = args.graph
    seasonal = args.seasonal
    fname = args.fname
    t0 = float(args.t0)
    t1 = float(args.t1)
    dt = float(args.dt)
    m  = int(t1/dt + 1.0e-6)+1

    #--- Read noise model parameters
    if os.path.exists(fname)==False:
        print('There is no {0:s}'.format(fname))
        sys.exit()
    try:
        fp_dummy = open('{0:s}'.format(fname),'r')
        results = json.load(fp_dummy)
        fp_dummy.close()
    except:
        print('Could not read estimatetrend.json')
        sys.exit()

    #--- Get list of noise model names
    noisemodels = results['NoiseModel']

    #--- Create dummy control file
    with open('dummy.ctl','w') as fp:
        fp.write('NoiseModels  ')
        for noisemodel in noisemodels.keys():
            fp.write(' {0:s}'.format(noisemodel))
        fp.write('\n')
        fp.write('Verbose    yes\n')
   
    #--- To get pretty figures, get units
    physical_unit = results['PhysicalUnit']
    time_unit = results['TimeUnit']
    if not time_unit=='days' and seasonal==True:
        print('Cannot add seasonal signal when time unit is not days')
        sys.exit()

    #--- Read control parameters into dictionary (singleton class)
    control = Control('dummy.ctl')

    try:
        verbose = control.params['Verbose']
    except:
        verbose = True

    if verbose==True:
        print("\n***************************************")
        print("    predict error, version 0.1.2")
        print("***************************************")

    #--- Get Classes
    white = White()
    powerlaw = Powerlaw()
    ggm = GGM()
    varyingannual = VaryingAnnual()
    ar1 = AR1()
    ammargrag = AmmarGrag()

    #--- Driving noise
    try:
        driving_noise = results['driving_noise']
    except:
        print('Could not find driving_noise')
        sys.exit()

    #--- Create empty autocovariance array
    t = np.zeros(m)
    if seasonal==True:
        H = np.ones((m,4))
    else:
        H = np.ones((m,2))
    F = np.zeros((m,0))
    x = np.zeros(m)

    #--- A noise model never has more than 2 parameters, to be safe -> 4
    param = [0.0]*4

    #--- extract parameter values
    for noisemodel in noisemodels.keys():
        if noisemodel=='White':
            fraction = noisemodels['White']['fraction']
            t_part,k_new = white.create_t(m,0,param)
        elif noisemodel=='Powerlaw':
            param[0] = noisemodels['Powerlaw']['kappa']
            fraction = noisemodels['Powerlaw']['fraction']
            t_part,k_new = powerlaw.create_t(m,0,param)
        elif noisemodel=='FlickerGGM' or noisemodel=='RandomWalkGGM' or \
                                                            noisemodel=='GGM':
            param[0] = noisemodels['GGM']['kappa']
            param[1] = noisemodels['GGM']['1-phi']
            fraction = noisemodels['GGM']['fraction']
            t_part,k_new = ggm.create_t(m,0,param)
        elif noisemodel=='VaryingAnnual':
            param[0] = noisemodels['VaryingAnnual']['phi']
            fraction = noisemodels['VaryingAnnual']['fraction']
            t_part,k_new = varyingannual.create_t(m,0,param)
        elif 'AR1' in noisemodels:
            param[0] = noisemodels['AR1']['phi']
            fraction = noisemodels['AR1']['fraction']
            t_part,k_new = ar1.create_t(m,0,param)
        else:
            print('Unknown noise model: {0:s}'.format(noisemodel))
            sys.exit()

        t += fraction * t_part
       
    #--- Multiply with driving noise
    t *= pow(driving_noise,2.0)

    #--- Measurement uncertainty based on noise model
    print('\nIf you want to put error bar on each measurement:')
    print('sigma = {0:.2f} {1:s}^2'.format(math.sqrt(t[0]),physical_unit))

    tt = 0.0
    k  = 0
    while tt<t0:
        if seasonal==True:
            H[k,2] = math.cos(2.0*math.pi/365.25 * tt)
            H[k,3] = math.sin(2.0*math.pi/365.25 * tt)
        tt += dt
        k += 1

    epochs = []
    trend_sigma = []
    while tt<t1:
        H[0:k,1] = np.linspace(-tt/2,tt/2,k)
        if seasonal==True:
            H[k,2] = math.cos(2.0*math.pi/365.25 * tt)
            H[k,3] = math.sin(2.0*math.pi/365.25 * tt)

        [theta,C_theta,ln_det_C,sigma_eta] = \
            ammargrag.compute_leastsquares(t,H[0:k,:],x[0:k],F[0:k,:],False) 

        if time_unit=='days':
            epochs.append(tt/365.25)
            trend_sigma.append(math.sqrt(C_theta[1,1]) * 365.25)
        elif time_unit=='seconds':
            epochs.append(tt/3600.0)
            trend_sigma.append(math.sqrt(C_theta[1,1]) * 3600.0)
        else:
            print('Unknown time unit: {0:s}'.format(time_unit))
            sys.exit()

        #--- prepare next epoch
        tt += dt
        k += 1

 
    if graph==True:
        fig = plt.figure(figsize=(5, 4), dpi=150)
        plt.plot(epochs,trend_sigma, label='trend sigma')
        if time_unit=='days':
            plt.xlabel('observation span [yr]')
            plt.ylabel('trend sigma [{0:s}/yr]'.format(physical_unit))
        else:
            plt.xlabel('observation span [h]')
            plt.ylabel('trend sigma [{0:s}/h]'.format(physical_unit))
        plt.legend()
        plt.show()


    #--- Write uncertainty to file
    fp = open('trend_sigma.out','w')
    fp.write('# predicted trend uncertainty\n')
    if time_unit=='days':
        fp.write('#\n#\n 1 time in years')
        fp.write('#\n 2 trend uncertainty in {0:s}/yr'.format(physical_unit))
    else:
        fp.write('#\n#\n 1 time in hours')
        fp.write('#\n 2 trend uncertainty in {0:s}/hr'.format(physical_unit))
    fp.write('#--------------------------------------------\n')
    for i in range(0,len(epochs)):
        fp.write('{0:e}  {1:e}\n'.format(epochs[i],trend_sigma[i]))
    fp.close()





import numpy
from pyfft.cuda import Plan
import pycuda.autoinit
import pycuda.gpuarray as gpuarray

# w,h,k are the array dimensions in a power of 2
# im1, im2 are the input 3d arrays of dtype complex64

w = h = k = 512
im1 = numpy.random.rand(w,h,k).astype(numpy.complex64)
im2 = numpy.random.rand(w,h,k).astype(numpy.complex64)

plan = Plan((w,h,k), normalize=True)

# forward transform on device
im1_gpu = gpuarray.to_gpu(im1)
plan.execute(im1_gpu)
im1_ft = im1_gpu.get()
del im1_gpu

im2_gpu = gpuarray.to_gpu(im2)
plan.execute(im2_gpu)
im2_ft = im2_gpu.get()
del im2_gpu

# do multiplication on host - can be done on device.
conv = im1_ft * im2_ft

#inverse transform on device
conv_gpu = gpuarray.to_gpu(conv)
del conv
plan.execute(conv_gpu, inverse=True)
corr_gpu = conv_gpu.get()

# Reference calculation on CPU:
im1_ft = numpy.fft.fftn(im1)
im2_ft = numpy.fft.fftn(im2)
conv = im1_ft * im2_ft
del im1
del im2
del im1_ft
del im2_ft
corr_cpu = numpy.fft.ifftn(conv)

print numpy.linalg.norm(corr_cpu - corr_gpu) / numpy.linalg.norm(corr_gpu)
